from PyQt6.QtWidgets import QLabel
from PyQt6.QtGui import QPixmap
from trilent.Utility import Reloader, V, Misc
from trilent.Widgets import PositionTypes


class Image(Misc):
    def __init__(self,
                 parent,
                 # Special
                 image_path  : str,
                 width       : int | str = None,
                 height      : int | str = None,
                 auto_size   : bool      = None,  # Controls if you don't want to enter width and height

                 # Stylesheet
                 background_color: str = None):

        # Reloader setup
        def image_func(v):
            pixmap = QPixmap(v)
            if self._reloader.cp['auto_size']:
                self._widget.adjustSize()

            size = self._reloader.process('width', self._reloader.cp['width']), self._reloader.process('height', self._reloader.cp['height'])

            pixmap = pixmap.scaled(*size)
            self._widget.setPixmap(pixmap)

        self._reloader = Reloader(parent.get_dpi(),
                                  setup_properties  = {'background_color': background_color, 'auto_size': auto_size,   'image_path': image_path, 'parent': parent,   'width': width,          'height': height},
                                  process_types     = {'background_color': 'color',          'auto_size': 'super raw', 'image_path': 'raw',      'parent': None,     'width': 'px-value;int', 'height': 'px-value;int'},
                                  default_values    = {'background_color': 'transparent',    'auto_size': True,        'image_path': None,       'parent': None,     'width': '1 inch',       'height': '1 inch'},
                                  property_types    = {'background_color': 'stylesheet',     'auto_size': 'access',    'image_path': 'special',  'parent': 'access', 'width': 'special',      'height': 'special'},
                                  special_functions = {'width': lambda v: self._widget.setGeometry(self.x, self.y, v, self.height), 'height': lambda v: self._widget.setGeometry(self.x, self.y, self.width, v), 'image_path': image_func},
                                  base              = f"background-color: {V}background_color{V}")

        # Protected widget specific members
        self._position_self = self._reloader.cp['parent']._position_children
        self._position_children = PositionTypes.NONE
        self._dpi = parent.get_dpi()

        # Qt setup
        self._widget = QLabel(self._reloader.cp['parent']._get_holder())
        self._widget.setStyleSheet(self._reloader.reload())

        # Check if Box's parent is also a box
        if self._reloader.cp['parent']._position_children == PositionTypes.BOX:
            self._reloader.cp['parent']._widget_box_add(self)



from PyQt6.QtWidgets import QGraphicsView, QGraphicsScene, QGraphicsRectItem
from PyQt6.QtGui import QPen, QColor, QBrush, QPainter
from PyQt6.QtCore import Qt, QRect
from trilent.Utility import Reloader, V, Misc
from trilent.Widgets import PositionTypes


class Shape:
    def __init__(self, qt_item: QGraphicsRectItem):
        self._qt_item = qt_item

    @property
    def x(self): return self._qt_item.x()
    @property
    def y(self): return self._qt_item.y()
    @property
    def width(self): return self._qt_item.rect().width()
    @property
    def height(self): return self._qt_item.rect().height()
    @property
    def scale(self): return self._qt_item.scale()
    @property
    def opacity(self): return self._qt_item.opacity()

    def set_scale(self, scale): return self._qt_item.setScale(scale)
    def set_opacity(self, opacity): return self._qt_item.setOpacity(opacity)

    def set_position(self, x, y):
        self._qt_item.setX(x)
        self._qt_item.setY(y)

    def set_size(self, width, height):
        self._qt_item.setRect(QRect(self.x, self.y, width, height))
        self._qt_item.scale()


class Canvas(Misc):
    def __init__(self,
                 parent,
                 # Special
                 width      : int | str = None,
                 height     : int | str = None,

                 # Stylesheet
                 background_color: str = None,
                 border_color    : str = None,
                 border_width    : str | int = None):

        # Reloader setup
        base = f"""QGraphicsView {{ background-color: {V}background_color{V}; border: {V}border_width{V} solid {V}border_color{V}; }}"""

        self._reloader = Reloader(parent.get_dpi(),
                                  setup_properties  = {'background_color': background_color,   'border_color': border_color,       'border_width': border_width,  'parent': parent,   'width': width,          'height': height},
                                  process_types     = {'background_color': 'color',            'border_color': 'color',            'border_width': 'px-value',    'parent': None,     'width': 'px-value;int', 'height': 'px-value;int'},
                                  default_values    = {'background_color': '#272727',          'border_color': '255, 255, 255, 3', 'border_width': 0,             'parent': None,     'width': '5 inch',       'height': '5 inch'},
                                  property_types    = {'background_color': 'stylesheet',       'border_color': 'stylesheet',       'border_width': 'stylesheet',  'parent': 'access', 'width': 'special',      'height': 'special'},
                                  special_functions = {'width': lambda v: self._widget.setGeometry(self.x, self.y, v, self.height), 'height': lambda v: self._widget.setGeometry(self.x, self.y, self.width, v)},
                                  base              = base)

        # Protected widget specific members
        self._position_self = self._reloader.cp['parent']._position_children
        self._position_children = PositionTypes.NONE
        self._dpi = parent.get_dpi()

        # Qt setup
        self._scene = QGraphicsScene()
        self._widget = QGraphicsView(self._scene, self._reloader.cp['parent']._get_holder())

        # Setup properties
        self._widget.setFixedSize(width, height)
        self._scene.setSceneRect(0,0, width, height)
        self._widget.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self._widget.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self._widget.setStyleSheet(self._reloader.reload())

        # Check if Canvas's parent is also a box
        if self._reloader.cp['parent']._position_children == PositionTypes.BOX:
            self._reloader.cp['parent']._widget_box_add(self)

    def square(self, x, y, width, height, color='white', border_color='black', border_width=1):
        rect = self._scene.addRect(x, y, width, height, QPen(QColor(border_color), border_width), QBrush(QColor(color)))
        return rect

    def circle(self, x, y, width, height, color='white', border_color='black', border_width=1):
        circle = self._scene.addEllipse(x, y, width, height, QPen(QColor(border_color), border_width), QBrush(QColor(color)))
        return Shape(circle)

    def line(self, x1, y1, x2, y2, color='white', width=1):
        line = self._scene.addLine(x1, y1, x2, y2, QPen(QColor(color), width))
        return Shape(line)

    def text(self, x, y, text, color='white', font=None):
        text_item = self._scene.addText(text)
        text_item.setDefaultTextColor(QColor(color))
        text_item.setPos(x, y)
        if font:
            text_item.setFont(font)
        return Shape(text_item)

    def clear(self):
        self._scene.clear()

    def set_size(self, width, height):
        self._widget.setFixedSize(width, height)
        self._scene.setSceneRect(0, 0, width, height)

    def set_pixel(self, x, y, color):
        self._scene.addRect(x, y, 1, 1, QPen(Qt.PenStyle.NoPen), QBrush(QColor(color)))



from .color import get_as_qt
from .unit import get_in_pixels
from .orientation import qt_orientation
from typing import Any, Literal


V = '␟'


class Reloader:
    def __init__(self, dpi: int,
                 setup_properties: dict[str, Any],
                 process_types: dict[str, str],
                 default_values: dict[str, Any],
                 property_types: dict[str, Literal['stylesheet', 'special', 'access']],
                 special_functions: dict[str, Any],
                 base: str):
        self.initial_parameters = {'setup_properties': setup_properties,
                                   'process_types': process_types,
                                   'default_values': default_values,
                                   'property_types': property_types,
                                   'special_functions': special_functions,
                                   'base': base}

        self.dpi = dpi
        self.property_types = property_types
        self.all_properties = list(setup_properties.keys())
        self.base = base

        self.process_types = process_types
        self.special_functions = special_functions

        self.cp = {}

        for p, v in setup_properties.items():
            self.cp[p] = default_values[p] if v is None else v

    def process(self, property_name: str, value: str):
        match self.process_types[property_name]:
            case 'color':
                return str(get_as_qt(value))
            case 'px-value':
                return str(get_in_pixels(value, self.dpi)) + 'px'
            case 'raw':
                return str(value)
            case 'px-value;int':
                return int(get_in_pixels(value, self.dpi))
            case 'orientation':
                return qt_orientation(value)
            case 'super raw':
                return value
            case None:
                return value

    def alternate_process(self, property_name: str, value: str):
        match property_name:
            case 'color':
                return str(get_as_qt(value))
            case 'px-value':
                return str(get_in_pixels(value, self.dpi)) + 'px'
            case 'raw':
                return str(value)
            case 'px-value;int':
                return int(get_in_pixels(value, self.dpi))
            case 'orientation':
                return qt_orientation(value)
            case 'super raw':
                return value
            case None:
                return value

    def reload(self):
        base = self.base

        for p, v in self.cp.items():
            if self.property_types[p] == 'special':
                self.special_functions[p](self.process(p, v))
            elif self.property_types[p] == 'stylesheet':
                base = base.replace(f"{V}{p}{V}", self.process(p, v))

        return base

    def set(self, p, v):
        if self.process_types[p] == 'special':
            self.special_functions[p](self.process(p, v))
        else:
            self.cp[p] = v

    def get(self, p):
        return self.cp[p]


def bench():
    for i in range(500):
        reloader.reload()


if __name__ == '__main__':
    reloader = Reloader(92,
                              {'parent': None, 'width': None, 'height': None, 'background_color': None,
                               'corner_roundness': None, 'alignment': None,
                               'side_alignment': None, 'wrap': None, 'gap': None,
                               'vertical_gap': None},
                              {'parent': None, 'width': 'px-value;int', 'height': 'px-value;int',
                               'background_color': 'color', 'corner_roundness': 'px-value', 'alignment': None,
                               'side_alignment': None, 'wrap': None, 'gap': 'px-value;int',
                               'vertical_gap': 'px-value;int'},
                              {'parent': None, 'width': '3 inch', 'height': '2 inch',
                               'background_color': 'cornflowerblue', 'corner_roundness': '10 px', 'alignment': 'start',
                               'side_alignment': 'start', 'wrap': True, 'gap': '1 px', 'vertical_gap': '1 px'},
                              {'parent': 'access', 'width': 'special', 'height': 'special',
                               'background_color': 'stylesheet', 'corner_roundness': 'stylesheet',
                               'alignment': 'access', 'side_alignment': 'access', 'wrap': 'access', 'gap': 'access',
                               'vertical_gap': 'access'},
                              {'width': lambda v: 0, 'height': lambda v: 0},
                              f"background-color: {V}background_color{V}; border-radius: {V}corner_roundness{V}")

    import cProfile
    cProfile.run('bench()')



from PyQt6.QtWidgets import QFrame
from trilent.Utility import Reloader, V, Misc, PositionTypes


class Widget(Misc):
    def __init__(self,
                 parent,
                 width: int | str,
                 height: int | str,
                 widget_color: str = None,
                 corner_roundness: str | int = None):

        # Stylesheet manager
        self._reloader = Reloader(parent.get_dpi(),
                                  setup_properties  = {'parent': parent,   'width': width,          'corner_roundness': corner_roundness, 'height': height,         'widget_color': widget_color},
                                  process_types     = {'parent': None,     'width': 'px-value;int', 'corner_roundness': 'px-value',       'height': 'px-value;int', 'widget_color': 'color'},
                                  default_values    = {'parent': None,     'width': '3 inch',       'corner_roundness': '0.03 inch',      'height': '2 inch',       'widget_color': 'cornflowerblue'},
                                  property_types    = {'parent': 'access', 'width': 'special',      'corner_roundness': 'stylesheet',     'height': 'special',      'widget_color': 'stylesheet'},
                                  special_functions = {'width': lambda v: self.set_size(v, self.height), 'height': lambda v: self.set_size(self.width, v)},
                                  base              = f"background-color: {V}widget_color{V}; border-radius: {V}corner_roundness{V};")

        self._position_self = self._reloader.cp['parent']._position_children
        self._position_children = PositionTypes.NONE

        self._widget = QFrame(parent._get_holder())
        self._widget.setStyleSheet(self._reloader.reload())

        if self._reloader.cp['parent']._position_children == PositionTypes.BOX:
            parent._widget_box_add(self)



__all__ = ['PositionTypes', 'Box',
           'Widget', 'Text',
           'Button', 'Entry',
           'Slider', 'CheckBox',
           'Switch', 'Canvas',
           'Image', 'Spacer',
           'Frame']

from .widget import PositionTypes, Widget
from .box import Box
from .text import Text
from .button import Button
from .entry import Entry
from .slider import Slider
from .check_box import CheckBox
from .switch import Switch
from .canvas import Canvas
from .image import Image
from .spacer import Spacer
from .frame import Frame

_ = PositionTypes, Box, Widget, Text, Button, Entry, Slider, CheckBox, Switch, Canvas, Image, Spacer, Frame



__all__ = ['get_in_pixels', 'get_as_qt', 'PropertyManager', 'Reloader',
           'qt_orientation', 'V', 'Misc', 'PositionTypes']

from .unit import get_in_pixels
from .color import get_as_qt
from .property_manager import PropertyManager
from .qss_reloader import Reloader, V
from .orientation import qt_orientation
from .misc import Misc, PositionTypes

_ = get_in_pixels, get_as_qt, PropertyManager, Reloader, V, qt_orientation, Misc, PositionTypes



from PyQt6.QtWidgets import QFrame
from trilent.Utility import Reloader, V, Misc
from trilent.Widgets import PositionTypes


class Frame(Misc):
    def __init__(self,
                 parent,
                 # Utility
                 width: int | str = None,
                 height: int | str = None,

                 # Styling
                 # Color
                 frame_color: str = None,
                 # Fancy
                 corner_roundness: int | str = None):

        def width_func(v): self._widget.setGeometry(self.x, self.y, v, self.height)
        def height_func(v): self._widget.setGeometry(self.x, self.y, self.width, v)

        # Reloader setup
        self._reloader = Reloader(parent.get_dpi(),
                                  setup_properties  = {'parent': parent,   'width': width,          'height': height,         'frame_color': frame_color,      'corner_roundness': corner_roundness},
                                  process_types     = {'parent': None,     'width': 'px-value;int', 'height': 'px-value;int', 'frame_color': 'color',          'corner_roundness': 'px-value'},
                                  default_values    = {'parent': None,     'width': '3 inch',       'height': '2 inch',       'frame_color': '255,255,255,30', 'corner_roundness': '0.03 inch'},
                                  property_types    = {'parent': 'access', 'width': 'special',      'height': 'special',      'frame_color': 'stylesheet',     'corner_roundness': 'stylesheet'},
                                  special_functions = {'width': width_func, 'height': height_func},
                                  base              = f"background-color: {V}frame_color{V}; border-radius: {V}corner_roundness{V}")

        # Protected widget specific members
        # noinspection PyProtectedMember
        self._position_self = self._reloader.cp['parent']._position_children
        self._position_children = PositionTypes.PLACE
        self._dpi = parent.get_dpi()

        # Qt setup
        # noinspection PyProtectedMember
        self._widget = QFrame(self._reloader.cp['parent']._get_holder())

        # Setup properties
        self._widget.setStyleSheet(self._reloader.reload())

        # Check if Box's parent is also a box
        # noinspection PyProtectedMember
        if self._reloader.cp['parent']._position_children == PositionTypes.BOX:
            # noinspection PyProtectedMember
            self._reloader.cp['parent']._widget_box_add(self)

    def _get_holder(self): return self._widget

    def get_dpi(self): return self._dpi



from .list_generator import generate_lists
from .initials_generator import generate_initials
from .easing_function_generator import generate_easing_function


class Animation:
    def __init__(self,
                 widget, name, end_value,
                 time: float = 0.3, curve='ease in out',
                 mode: str = "",
                 animation_finished=lambda: ...):
        # Initials
        self.widget = widget
        self.time = time
        self.name = name
        self.animation_finished = animation_finished
        self.mode = mode
        self.elapsed_time = 0

        self.window = widget.get_top_parent()

        self.start_value, self.end_value, self._type = generate_initials(name, end_value, widget)

        self._current_value = self.start_value

        self.easing_function = generate_easing_function(curve)
        self.listings = generate_lists(name, self._type, self.start_value, self.end_value, mode)

        if self.name == 'x':
            self.setter = lambda v: self.widget.set_position(self._current_value, self.widget.y)
        elif self.name == 'y':
            self.setter = lambda v: self.widget.set_position(self.widget.x, self._current_value)
        else:
            self.setter = lambda v: self.widget.set(self.name, self._current_value)

        self._after_function = None
        self._stop_elapsed_time = 0
        self._is_stopping = False
        self._stop_duration = 1

    def start(self):
        if self.start_value != self.end_value:
            self.window.add_update_function(self.update, parse_delta=True)
        else:
            self.animation_finished()

    def update(self, delta):
        if self._is_stopping:
            self._stop_elapsed_time += delta
            proportion = self.easing_function(min(self._stop_elapsed_time / self._stop_duration, 1))
            self.time += proportion * delta

            if self._stop_elapsed_time >= self._stop_duration:
                self._is_stopping = False
                self.window.remove_update_function(self.update)
                self.animation_finished()
                self._after_function()
                return

        self.elapsed_time += delta
        proportion = self.easing_function(min(self.elapsed_time / self.time, 1))

        index = int(proportion * (len(self.listings) - 1))
        index = max(min(index, len(self.listings) - 1), 0)

        self._current_value = self.listings[index]

        self.setter(self._current_value)

        if self._current_value == self.listings[-1]:
            self.window.remove_update_function(self.update)

            if self.name in ['x', 'y']:
                self.widget.set_position(self.end_value, self.widget.y) if self.name == 'x' else self.widget.set_position(self.widget.x, self._current_value)
            else:
                self.widget.set(self.name, self.end_value if not isinstance(self.end_value, str) else self.end_value.removesuffix('px'))

            self.animation_finished()

    def stop(self, stop_duration=1, after_function=lambda: ...):
        self._is_stopping = True
        self._stop_duration = stop_duration
        self._stop_elapsed_time = 0
        self._after_function = after_function

    def change(self, new_end_value=None, break_speed=None, other_speed=None):
        if new_end_value is None:
            new_end_value = self._current_value
        if break_speed is None:
            break_speed = 0.1
        if other_speed is None:
            other_speed = self.time - self.elapsed_time

        self.stop(break_speed)

        break_speed = max(break_speed, 0.1)

        self.window.after(int((break_speed * 1000)) - 100, lambda: Animation(widget=self.widget, name=self.name, end_value=new_end_value, time=other_speed, curve=self.easing_function, animation_finished=self.animation_finished).start())




# This is mainly using by the Text widget.
from PyQt6.QtCore import Qt
from typing import Literal

QT_DICT = {
    'top left':      Qt.AlignmentFlag.AlignTop     | Qt.AlignmentFlag.AlignLeft,
    'top center':    Qt.AlignmentFlag.AlignTop     | Qt.AlignmentFlag.AlignHCenter,
    'top right':     Qt.AlignmentFlag.AlignTop     | Qt.AlignmentFlag.AlignRight,
    'center left':   Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignLeft,
    'center center': Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignHCenter,
    'center right':  Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignRight,
    'bottom left':   Qt.AlignmentFlag.AlignBottom  | Qt.AlignmentFlag.AlignLeft,
    'bottom center': Qt.AlignmentFlag.AlignBottom  | Qt.AlignmentFlag.AlignHCenter,
    'bottom right':  Qt.AlignmentFlag.AlignBottom  | Qt.AlignmentFlag.AlignRight,

    'center': Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignHCenter,
    'top'   : Qt.AlignmentFlag.AlignTop     | Qt.AlignmentFlag.AlignHCenter,
    'bottom': Qt.AlignmentFlag.AlignBottom  | Qt.AlignmentFlag.AlignHCenter,
    'left'  : Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignLeft,
    'right' : Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignRight
}


QT_DICT_LITERAL = Literal[
    'top left',
    'top center',
    'top right',
    'center left',
    'center center',
    'center right',
    'bottom left',
    'bottom center',
    'bottom right',
    'center',
    'top',
    'bottom',
    'left',
    'right']


def qt_orientation(name: QT_DICT_LITERAL):
    return QT_DICT[name]





def generate_initials(name, end_value, widget):
    if name in ['x', 'y']:
        start_value = getattr(widget, name)
        end_value = widget._reloader.alternate_process('px-value;int', end_value)
        return start_value, end_value, 'px-value;int'
    elif name == 'text':
        start_value = widget.get(name)
        return start_value, end_value, 'text'
    elif isinstance(end_value, str) and end_value.__contains__(' -> '):
        start_value = widget.get(name)
        if not start_value.__contains__(' -> '):
            start_value = f"{start_value} -> " * end_value.count(' -> ')
            start_value = start_value.removesuffix(' -> ')

        return start_value, end_value, 'gradient color'
    else:
        start_value = widget._reloader.process(name, widget._reloader.cp[name])
        end_value = widget._reloader.process(name, end_value)
        type = widget._reloader.initial_parameters['process_types'][name]
        return start_value, end_value, type




from trilent.Widgets import Slider
from typing import Literal
from PyQt6.QtCore import Qt, QEvent
from PyQt6.QtWidgets import QSlider


class Switch(Slider):
    def __init__(self,
                 parent,
                 # Special
                 width      : int | str = '0.45 inch',
                 height     : int | str = None,
                 command    : lambda: ... = None,

                 # Access
                 modes      : str | int = 2,

                 # Stylesheet
                 unfilled_color  : str = None,
                 button_color    : str = None,
                 filled_color    : str = None,

                 orientation            : Literal['horizontal', 'vertical'] = 'horizontal',
                 corner_roundness       : str | int = None,
                 handle_corner_roundness: str | int = None,

                 # Kick starters
                 starting_state         : int = 1):

        if handle_corner_roundness is None:
            handle_corner_roundness = 15 // 2

        super().__init__(parent, width, height, command, 1, modes, 1, unfilled_color, button_color, filled_color, orientation, corner_roundness=corner_roundness, handle_corner_roundness=handle_corner_roundness)

        self._state = starting_state
        self._modes = modes

        self._moveHandled = False

        # Make it behave like a toggler instead of a picker
        self._widget.mousePressEvent = self.mousePressEvent
        self._widget.mouseReleaseEvent = self.mouseReleaseEvent
        self._widget.mouseMoveEvent = self.mouseMoveEvent

    def mousePressEvent(self, event):
        # Ignore mouse press event to prevent the handle from being moved
        if event.button() == Qt.MouseButton.LeftButton:
            event.ignore()
            return
        super(QSlider, self._widget).mousePressEvent(event)

    @staticmethod
    def find_closest_number(array, target):
        closest_num = array[0]
        min_diff = abs(array[0] - target)

        for num in array:
            diff = abs(num - target)
            if diff < min_diff:
                min_diff = diff
                closest_num = num

        return closest_num

    def mouseMoveEvent(self, event: QEvent):
        x = event.pos().x()
        width = self._widget.width()
        new_value = (x / width) * (self._widget.maximum() - self._widget.minimum()) + self._widget.minimum()

        width_mappings = [i for i in range(((self._widget.maximum() - self._widget.minimum()) + self._widget.minimum()) + 1)]

        closest_value = self.find_closest_number(width_mappings, new_value)

        self._widget.setValue(int(closest_value))

        self._moveHandled = True

        event.accept()

    def mouseReleaseEvent(self, event):
        if not self._moveHandled:
            if event.button() == Qt.MouseButton.LeftButton:
                # Increment the state and wrap around if it exceeds the modes
                self._state = self._widget.value() + 1
                if self._state > self._widget.maximum():
                    self._state = self._widget.minimum()
                self._widget.setValue(self._state)

            super(QSlider, self._widget).mouseReleaseEvent(event)
        else:
            self._moveHandled = False



from .general import fallback, CACHING_LIMIT, FUNCTIONS, center_y_positions
from functools import lru_cache


@lru_cache(CACHING_LIMIT // FUNCTIONS)
def calculate_center_positions(widths: tuple[int, ...], heights: tuple[int, ...], flex_box_size: tuple[int | str, int | str], wrap: bool, gap: int, vertical_gap: int, side_alignment: str) -> list[list[int]]:
    if side_alignment == 'end':
        x, y = 0, flex_box_size[1]
    else:
        x, y = 0, 0

    positions: list[list[int]] = []
    column_heights: list[int] = []
    row_start_index = 0
    wraps = 1
    total_height = 0

    for i, width in enumerate(widths):
        if wrap and x + width > flex_box_size[0] and i != 0:
            # Center the previous row
            row_width = x - gap
            shift = (flex_box_size[0] - row_width) // 2
            for j in range(row_start_index, i):
                positions[j][0] += shift

            # Start new row
            row_start_index = i
            x = 0
            if side_alignment == 'end':
                y -= max(column_heights) + vertical_gap
            else:
                y += max(column_heights) + vertical_gap
            total_height += max(column_heights) + vertical_gap

            wraps += 1
            column_heights = []

        if side_alignment == 'end':
            positions.append([x, y - heights[i]])
        else:
            positions.append([x, y])

        column_heights.append(heights[i])
        x += width + gap

    # Center the last row
    if row_start_index < len(widths):
        row_width = x - gap
        shift = (flex_box_size[0] - row_width) // 2
        for j in range(row_start_index, len(widths)):
            positions[j][0] += shift

    total_height += max(column_heights) if column_heights else 0

    # Fallback is used to fix a bug, You can ignore this.
    if wraps >= len(widths):
        positions = fallback(positions, heights, vertical_gap, side_alignment, flex_box_size[1])

    if side_alignment == 'center':
        positions = center_y_positions(positions, flex_box_size, total_height)

    return positions




def samify(str1, str2, spacify=False):
    if spacify:
        str1 = str1.replace(' ', '⠀')

    if len(str1) < len(str2):
        str1 += ' ' * (len(str2) - len(str1))
    elif len(str1) > len(str2):
        str2 += ' ' * (len(str1) - len(str2))

    return str1, str2


def redundantify(string):
    return ' '.join(''.join(string).split()).rstrip()


def text_gradient(str1, str2, mode):
    spacify = False

    if " | " in mode:
        splitted = mode.split(' | ')
        spacify  = bool(splitted[-1].lower())
        mode     = ''.join(splitted[:-1])

    str1, str2 = samify(str1, str2, spacify)

    transitions = [str1]

    current = list(str1)
    goal = list(str2)
    center_index = len(str1) // 2
    size = len(str1) - 1

    if mode == 'wave' or mode == '':
        for i in range(max(len(str1), len(str2))):
            current[i] = goal[i]
            transitions.append(redundantify(current))
    elif mode == 'pulse':
        for i in range(max(len(str1), len(str2)) // 2):
            current[center_index - i] = goal[center_index - i]
            current[center_index + i] = goal[center_index + i]
            transitions.append(redundantify(current))
    elif mode == 'roll':
        for i in range(max(len(str1), len(str2))):
            current[size - i] = goal[size - i]
            transitions.append(redundantify(current))

    return transitions


if __name__ == '__main__':
    print(text_gradient('Test.', 'Secondary text.', 'pulse'))
